/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.robokind.demo.robot.replication;

import org.robokind.api.messaging.MessageSender;
import org.robokind.api.motion.protocol.RobotResponse;
import org.robokind.api.motion.messaging.RemoteRobot;
import org.robokind.api.motion.Robot;
import java.util.Arrays;
import java.util.Map;
import org.jflux.api.core.Adapter;
import org.jflux.api.core.config.Configuration;
import java.util.List;
import org.jflux.api.core.Listener;
import org.jflux.api.core.Source;
import org.jflux.api.core.chain.ListenerChain;
import org.jflux.api.core.event.Event;
import org.jflux.api.core.event.Header;
import org.jflux.api.core.node.ConsumerNode;
import org.jflux.api.core.node.DefaultConsumerNode;
import org.osgi.framework.BundleContext;
import org.robokind.api.common.lifecycle.ServiceLifecycleProvider;
import org.robokind.api.common.lifecycle.utils.ManagedServiceFactory;
import org.robokind.api.common.lifecycle.utils.SimpleLifecycle;
import org.robokind.api.common.osgi.lifecycle.OSGiComponentFactory;
import org.robokind.api.common.position.NormalizableRange;
import org.robokind.api.common.position.NormalizedDouble;
import org.robokind.api.motion.Joint;
import org.robokind.api.motion.JointProperty;
import org.robokind.api.motion.messaging.RemoteRobotHost;
import org.robokind.api.motion.protocol.RobotDefinitionResponse;
import static org.robokind.api.common.lifecycle.config.RKDependencyConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKLifecycleConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKManagedGroupConfigUtils.*;

/**
 *
 * @author matt
 */
public class RobotReplicator {
    
    public static void startRobotReplicator(
            BundleContext context, String replicationGroupId, 
            String connectionConfigId, String destConfigId, List<Joint> joints,
            long replicationPeriod, String[] propNames,
            ManagedServiceFactory fact){
        ReplicationUtils.startReplicator(
                context, replicationGroupId, 
                destConfigId, connectionConfigId, 
                2000L, replicationPeriod, fact);
        replicateJoints(replicationGroupId, replicationGroupId, joints, propNames, fact);
    }
    
    private static void replicateJoints(
            String replicationGroupId, String bufferRouteKeyPrefix,
            List<Joint> joints, String[] propNames, ManagedServiceFactory fact){
        for(Joint joint : joints){
            replicateJointProperties(
                    replicationGroupId, bufferRouteKeyPrefix, 
                    joint, propNames, fact);
        }
    }
    
    private static void replicateJointProperties(
            String replicationGroupId, String bufferRouteKeyPrefix,
            Joint joint, String[] propNames, ManagedServiceFactory fact){
        if(propNames == null){
            for(JointProperty p : joint.getProperties()){
            String bufferRouteKey = 
                    bufferRouteKeyPrefix + joint.getId().toString() + p.getPropertyName();
            addJointPropertyBuffer(replicationGroupId, bufferRouteKey, p, fact);
            }
            return;
        }
        for(String name : propNames){
            JointProperty p = joint.getProperty(name);
            if(p == null){
                continue;
            }
            String bufferRouteKey = 
                    bufferRouteKeyPrefix + joint.getId().toString() + name;
            addJointPropertyBuffer(replicationGroupId, bufferRouteKey, p, fact);
        }
    }
    
    private static void addJointPropertyBuffer(
            String replicationGroupId, String bufferRouteKey,
            JointProperty prop, ManagedServiceFactory fact){
        Class messageClass = NormalizedDouble.class;
        Source<NormalizedDouble> propSource = new JointPropertySource(prop);
        ReplicationUtils.createAlternateReplicationBuffer(
                replicationGroupId, bufferRouteKey, 
                propSource, messageClass, fact);
    }
    
    public static void startRobotReceiver(
            String receiverGroupId, String connectionConfigId, String destConfigId, 
            List<Joint> joints, String[] propNames, ManagedServiceFactory fact){
        ReplicationUtils.startReceiver(
                receiverGroupId, destConfigId, connectionConfigId, fact);
        receiveJoints(receiverGroupId, receiverGroupId, joints, propNames, fact);
    }
    
    private static void receiveJoints(
            String receiverGroupId, String bufferRouteKeyPrefix, 
            List<Joint> joints, String[] propNames, ManagedServiceFactory fact){
        for(Joint joint : joints){
            receiveJointProperties(
                    receiverGroupId, bufferRouteKeyPrefix, 
                    joint, propNames, fact);
        }
    }
    
    private static void receiveJointProperties(
            String receiverGroupId, String bufferRouteKeyPrefix, 
            Joint joint, String[] propNames, ManagedServiceFactory fact){
        if(propNames == null){
            for(JointProperty p : joint.getProperties()){
                String bufferRouteKey = 
                        bufferRouteKeyPrefix + joint.getId().toString() + p.getPropertyName();
                addJointPropertyRoute(receiverGroupId, bufferRouteKey, p, fact);
            }
            return;
        }
        for(String name : propNames){
            JointProperty p = joint.getProperty(name);
            if(p == null){
                continue;
            }
            String bufferRouteKey = 
                    bufferRouteKeyPrefix + joint.getId().toString() + name;
            addJointPropertyRoute(receiverGroupId, bufferRouteKey, p, fact);
        }
    }
    
    private static void addJointPropertyRoute(
            String receiverGroupId, String bufferRouteKey,
            JointProperty prop, ManagedServiceFactory fact){
        String consumerIdKey = "jointPropRoute";
        String consumerId = bufferRouteKey;
        launchJointPropConsumer(consumerIdKey, consumerId, prop, fact);
        ReplicationUtils.createReceiverRoute(receiverGroupId, bufferRouteKey, 
                consumerIdKey, consumerId, NormalizedDouble.class, fact);
    }
    
    private static void launchJointPropConsumer(
            String consumerIdKey, String consumerId, 
            JointProperty prop, ManagedServiceFactory fact){
        Listener<NormalizedDouble> listener = new JointPropertyListener(prop);
        ConsumerNode<Event<Header,NormalizedDouble>> consumer = 
                new DefaultConsumerNode<Event<Header,NormalizedDouble>>(
                        new ListenerChain<Event<Header, NormalizedDouble>, NormalizedDouble>(
                                new Event.EventDataAdapter<Header, NormalizedDouble>(), 
                                listener));
        consumer.start();
        ServiceLifecycleProvider lifecycle = new SimpleLifecycle(
                consumer, ConsumerNode.class, consumerIdKey, consumerId, null);
        fact.createService(lifecycle, null).start();
    }
    
    
    private final static String theRobotHost = "robotHost";
    private final static String theRemoteRobot = "remoteRobot";
    public final static String REPLICATION_CONNECTION_CONFIG_ID = "robotReplicationConnectionConfig";
    public final static String RECEIVER_CONNECTION_CONFIG_ID = "robotReceiverConnectionConfig";
    public final static String DESTINATION_CONFIG_ID = "robotReplicationDestinationConfig";
    
    public static <T> Configuration<String> buildRobotReplicationLifecycleConfig(
            final BundleContext context, final String connectConfigId, 
            final String destConfigId, final long replicationPeriod){
        List<Configuration<String>> deps = Arrays.asList(buildLifecycleDependencyConfig(
                theRobotHost, RemoteRobotHost.class, null, null)
        );
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{Object.class.getName()}, null, deps, 
                new Adapter<Map<String, Object>, Object>() {
                    @Override public Object adapt(Map<String, Object> a) {
                        RemoteRobotHost host = (RemoteRobotHost)a.get(theRobotHost);
                        final MessageSender<RobotResponse> sender = host.getResponseSender();
                        final Robot r = host.getRobot();
                        Listener<RobotResponse> l = new Listener<RobotResponse>() {
                            private boolean myRunFlag = true;
                                @Override
                                public void handleEvent(RobotResponse input) {
                                    if(myRunFlag && input instanceof RobotDefinitionResponse){
                                        startRobotReplicator(context, 
                                                r.getRobotId().getRobtIdString(), 
                                                connectConfigId, destConfigId, 
                                                r.getJointList(), replicationPeriod, 
                                                null, new OSGiComponentFactory(context));
                                        myRunFlag = false;
                                    }
                                }
                            };
                        sender.addListener(l);
                        return new Object();
                    }
                }));        
    }
    public static <T> Configuration<String> buildRobotReceiverLifecycleConfig(
            final BundleContext context, final String connectConfigId, 
            final String destConfigId){
        List<Configuration<String>> deps = Arrays.asList(buildLifecycleDependencyConfig(
                theRemoteRobot, RemoteRobot.class, null, null)
        );
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{Object.class.getName()}, null, deps, 
                new Adapter<Map<String, Object>, Object>() {
                    @Override public Object adapt(Map<String, Object> a) {
                        RemoteRobot r = (RemoteRobot)a.get(theRemoteRobot);
                        startRobotReceiver(r.getRobotId().getRobtIdString(), 
                                connectConfigId, destConfigId, 
                                ((Robot)r).getJointList(), null, 
                                new OSGiComponentFactory(context));
                        return new Object();
                    }
                }));        
    }
    
    static class JointPropertySource<T> implements Source<NormalizedDouble> {
        private JointProperty<T> myProperty;
        
        public JointPropertySource(JointProperty<T> property){
            if(property == null){
                throw new NullPointerException();
            }
            myProperty = property;
        }
        
        @Override
        public NormalizedDouble getValue() {
            T t = myProperty.getValue();
            NormalizableRange<T> range = myProperty.getNormalizableRange();
            if(t == null || range == null){
                return null;
            }
            return range.normalizeValue(t);
        }
    }
    
    static class JointPropertyListener<T> implements Listener<NormalizedDouble> {
        private JointProperty<T> myProperty;
        
        public JointPropertyListener(JointProperty<T> property){
            if(property == null){
                throw new NullPointerException();
            }
            myProperty = property;
        }

        @Override
        public void handleEvent(NormalizedDouble input) {
            if(!myProperty.getWriteable()){
                return;
            }
            NormalizableRange<T> range = myProperty.getNormalizableRange();
            if(input == null || range == null){
                return;
            }
            T t = range.denormalizeValue(input);
            if(t != null){
                myProperty.setValue(t);
            }
        }
    }
}
